from __future__ import annotations

import re
import time
from datetime import datetime
from random import random
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union

from ...logger import logger
from ..common import SqlProcessorAssertionError
from .base import Backend, Partition, Row, SaveMode, Table, TableMeta
from .sql_dialect import SqlDialect, SqlExpr
from .sql_dialect.bigquery import BqSqlDialect
from .sql_dialect.clickhouse import ChSqlDialect
from .sql_dialect.postgres import PgSqlDialect

__all__ = ["RdbBackend", "SqlExpr"]

from ...udf import udfs
from .base import Col

if TYPE_CHECKING:
    from sqlalchemy.engine import ResultProxy
    from sqlalchemy.engine.base import Connection, Engine
    from sqlalchemy.engine.reflection import Inspector
    from sqlalchemy.sql.elements import TextClause
    from sqlalchemy.types import TypeEngine


class TimeLog:
    time_took_tpl = "time took: {time_took:.3f}s"

    def __init__(self, start_log: str, end_log_tpl: str):
        self.start_log = start_log
        self.end_log_tpl = end_log_tpl
        self.start_dt = None

    def __enter__(self):
        logger.info(self.start_log)
        self.start_dt = datetime.now()

    def __exit__(self, exc_type, exc_val, exc_tb):
        time_took = (datetime.now() - self.start_dt).total_seconds()
        logger.info(self.end_log_tpl.format(time_took=time_took))


def _exec_sql(conn, sql: Union[str, TextClause, List[str]], *args, **kwargs) -> ResultProxy:
    from sqlalchemy.sql.elements import TextClause

    conn: Connection = conn
    with TimeLog(
        f"start to execute sql: {sql}, kwargs={kwargs}", f"end to execute sql({TimeLog.time_took_tpl}): {sql}"
    ):
        if isinstance(sql, (str, TextClause)):
            return conn.execute(sql, *args, **kwargs)
        else:
            execute_result = None
            for each_sql in sql:
                each_sql = each_sql.strip()
                if each_sql:
                    execute_result = conn.execute(each_sql, *args, **kwargs)
            return execute_result


def _quote_str(x):
    return f"'{x}'" if isinstance(x, str) else f"{x}"


class RdbTable(Table):
    def __init__(self, backend, sql: str, actions: List[Tuple] = None):
        self.backend: RdbBackend = backend
        self.db_config: SqlDialect = backend.sql_dialect
        self.sql = sql
        self._exec_sql = lambda sql: _exec_sql(self.backend.conn, sql)
        self._actions = actions or []

        self._temp_table_time_prefix = lambda: f"t_{round(time.time() * 1000)}_{int(random() * 100000):04d}"
        self._is_simple_query = lambda sql: re.match(r"^select \* from [\w.]+$", sql)
        self._table_name_of_simple_query = lambda sql: re.match(r"select \* from ([\w.]+)", sql).group(1)

    @staticmethod
    def from_table_meta(backend, table_meta: TableMeta):
        table = RdbTable(backend, f"select * from {table_meta.get_full_table_name(backend.temp_schema)}")
        if table_meta.has_partitions():
            for pt in table_meta.partitions:
                if pt.field not in table.field_names():
                    table = table.with_column(pt.field, backend.sql_expr.for_value(pt.value))
                else:
                    if pt.value is not None:
                        logger.warning(
                            f"partition column already exists in table {table_meta.table_name}, but right now we"
                            f" provided a new value {pt.value} for partition column {pt.field}. Will ignore it."
                        )
        return table

    def _execute_actions(self):
        for action in self._actions:
            if action[0] == "limit":
                count = action[1]
                with TimeLog(
                    f"start to execute action: {action}", f"end to execute action({TimeLog.time_took_tpl}): {action}"
                ):
                    prefix = f"{self._temp_table_time_prefix()}"
                    limit_result_table_name = f"{prefix}_limit_{count}"
                    if self._is_simple_query(self.sql):
                        temp_table_name = self._table_name_of_simple_query(self.sql)
                        self._exec_sql(
                            self.backend.sql_dialect.create_view_sql(
                                limit_result_table_name, f"select * from {temp_table_name} limit {count}"
                            )
                        )
                    else:
                        temp_table_name = f"{prefix}_limit_{count}_source"
                        self._exec_sql(self.backend.sql_dialect.create_view_sql(temp_table_name, self.sql))
                        self._exec_sql(
                            self.backend.sql_dialect.create_view_sql(
                                limit_result_table_name,
                                f"select * from {self.backend.temp_schema}.{temp_table_name} limit {count}",
                            )
                        )
                    self.sql = f"select * from {self.backend.temp_schema}.{limit_result_table_name}"
            elif action[0] == "newcol":
                name, value = action[1], action[2]
                with TimeLog(
                    f"start to execute action: {action}", f"end to execute action({TimeLog.time_took_tpl}): {action}"
                ):
                    prefix = self._temp_table_time_prefix()
                    # for pg: max table name chars allowed is 63, the max length is 55 for newcol_table_name
                    newcol_table_name = f"{prefix}_newcol_{name[:30]}"
                    if self._is_simple_query(self.sql):
                        temp_table_name = self._table_name_of_simple_query(self.sql)
                        field_names = self._field_names(temp_table_name)
                        select_sql = f'select {", ".join(field_names)}, {value} as {name} from {temp_table_name}'
                    else:
                        # for pg: max table name chars allowed is 63, the max length is 62 for newcol_table_name
                        temp_table_name = f"{prefix}_newcol_{name[:30]}_source"
                        self._exec_sql(self.db_config.create_view_sql(temp_table_name, self.sql))
                        field_names = self._field_names(f"{self.backend.temp_schema}.{temp_table_name}")
                        select_sql = (
                            f'select {", ".join(field_names)}, {value} as {name} from'
                            f" {self.backend.temp_schema}.{temp_table_name}"
                        )

                    self._exec_sql(self.db_config.create_view_sql(newcol_table_name, select_sql))
                    self.sql = f"select * from {self.backend.temp_schema}.{newcol_table_name}"
            else:
                raise SqlProcessorAssertionError(f"unsupported action: {action}")
        self._actions = []

    def is_empty(self) -> bool:
        return self.count() == 0

    def field_names(self) -> List[str]:
        self._execute_actions()
        if self._is_simple_query(self.sql):
            temp_table_name = self._table_name_of_simple_query(self.sql)
            return self._field_names(temp_table_name)
        prefix = f"{self._temp_table_time_prefix()}"
        field_names_result_table_name = f"{prefix}_field_names"
        self._exec_sql(self.backend.sql_dialect.create_view_sql(field_names_result_table_name, self.sql))
        return self._field_names(f"{self.backend.temp_schema}.{field_names_result_table_name}")

    def _field_names(self, table_name: str) -> List[str]:
        result: ResultProxy = self._exec_sql(f"select * from {table_name} limit 0")
        result.close()
        return list(result.keys())

    def first(self) -> RdbRow:
        all_action_are_limit = all([action[0] == "limit" for action in self._actions])
        if all_action_are_limit:
            min_limit = min([action[1] for action in self._actions]) if len(self._actions) > 0 else 1

            result: ResultProxy = self._exec_sql(self.sql)
            if min_limit <= 0:
                return RdbRow(list(result.keys()), None)
            with TimeLog(
                f"start to fetch first row: {self.sql}", f"end to fetch first row({TimeLog.time_took_tpl}): {self.sql}"
            ):
                row = result.first()
            return RdbRow(list(result.keys()), row)

        self._execute_actions()

        result: ResultProxy = self._exec_sql(self.sql)
        with TimeLog(
            f"start to fetch first row: {self.sql}", f"end to fetch first row({TimeLog.time_took_tpl}): {self.sql}"
        ):
            row = result.first()
        return RdbRow(list(result.keys()), row)

    def limit(self, count: int) -> RdbTable:
        return RdbTable(self.backend, self.sql, self._actions + [("limit", count)])

    def with_column(self, name: str, value: any) -> RdbTable:
        return RdbTable(self.backend, self.sql, self._actions + [("newcol", name, value)])

    def collect(self) -> List[RdbRow]:
        return self._collect()

    def _collect(self, row_count: int = None) -> List[RdbRow]:
        self._execute_actions()

        result: ResultProxy = self._exec_sql(self.sql)
        # collect at most 1000 rows for now
        max_rows = 1000 if row_count is None else row_count
        with TimeLog(
            f"start to fetch first row: {self.sql}", f"end to fetch first row({TimeLog.time_took_tpl}): {self.sql}"
        ):
            try:
                rows = result.fetchmany(max_rows)
            except Exception as e:
                print("result.fetchmany(max_rows) found error: ", e)
                if e.args[0] == "This result object does not return rows. It has been closed automatically.":
                    return []
                else:
                    raise e
        if row_count is None and len(rows) == max_rows:
            logger.warning(
                f"found {max_rows} items, but there may be more, will only fetch {max_rows} items at most for sql:"
                f" {self.sql}"
            )
        rows = [RdbRow(list(result.keys()), row) for row in rows]
        result.close()
        return rows

    def show(self, count: int = 20):
        keys = self.field_names()
        rows = self._collect(count)
        print("\t".join(keys))
        for row in rows:
            print("\t".join([_quote_str(item) for item in row]))

    def count(self) -> int:
        temp_table_name = self.resolve_to_temp_table()
        return self._exec_sql(f"select count(1) from {self.backend.temp_schema}.{temp_table_name}").first()[0]

    def resolve_to_temp_table(self) -> str:
        self._execute_actions()
        if self._is_simple_query(self.sql) and "." not in self._table_name_of_simple_query(self.sql):
            temp_table_name = self._table_name_of_simple_query(self.sql)
        else:
            prefix = self._temp_table_time_prefix()
            temp_table_name = f"{prefix}_count"
            self._exec_sql(self.db_config.create_view_sql(temp_table_name, self.sql))
        self.sql = f"select * from {self.backend.temp_schema}.{temp_table_name}"
        return temp_table_name

    def save_to_temp_table(self, name: str):
        temp_table_name = self.resolve_to_temp_table()
        if temp_table_name != name:
            if "." in name:
                raise SqlProcessorAssertionError(
                    "renaming should only happen in temp database, "
                    f"so name must be pure TABLE_NAME when renaming tables, found: {name}"
                )
            if temp_table_name != name:
                if self.backend.table_exists(TableMeta(name)):
                    raise SqlProcessorAssertionError(
                        "we are trying to replace an existing temp table, it is not supported right now. table name:"
                        f" {name}"
                    )
                self._exec_sql(
                    self.db_config.create_view_sql(name, f"select * from {self.backend.temp_schema}.{temp_table_name}")
                )

    def save_to_table(self, target_table: TableMeta):
        temp_table_name = self.resolve_to_temp_table()

        field_names = self.field_names()
        for pt in target_table.partitions:
            if pt.field not in field_names:
                raise Exception(
                    f"does not found partition field `{pt.field}` in source table for target table"
                    f" {target_table.table_name}, all fields are in source table: {field_names}"
                )

        cols = self.backend.get_columns(temp_table_name, self.backend.temp_schema)
        if self.backend.table_exists(target_table):
            cols = self.backend.get_columns(target_table.pure_table_name, target_table.dbname)
        else:
            db = target_table.table_name[: target_table.table_name.index(".")]
            self._exec_sql(self.db_config.create_db_sql(db))
            self._exec_sql(
                self.db_config.create_table_with_partitions_sql(target_table.table_name, cols, target_table.partitions)
            )

        target_table_name = target_table.get_full_table_name(self.backend.temp_schema)

        partitions_to_save = self._get_save_partitions(target_table, temp_table_name)

        if not self.db_config.create_partition_automatically():
            source_table_name = f"{self.backend.temp_schema}.{temp_table_name}"
            sqls = self.db_config.create_partitions_with_data_sqls(
                source_table_name, target_table_name, [col["name"] for col in cols], partitions_to_save
            )
            for sql in sqls:
                self._exec_sql(sql)
        else:
            cols = [col["name"] for col in cols]
            col_names = ", ".join(cols)
            converted_col_names = ", ".join(
                self.db_config.convert_pt_col_expr(cols, [pt.field for pt in target_table.partitions])
            )
            if not partitions_to_save:
                self._exec_sql(
                    self.db_config.insert_data_sql(
                        target_table_name,
                        col_names,
                        f"select {converted_col_names} from {self.backend.temp_schema}.{temp_table_name}",
                        [],
                    )
                )
            for partitions in partitions_to_save:
                filter_expr = " and ".join(
                    [f"{pt.field} = {self.backend.sql_expr.for_value(pt.value)}" for pt in partitions]
                )
                self._exec_sql(
                    self.db_config.insert_data_sql(
                        target_table_name,
                        col_names,
                        f"select {converted_col_names} from {self.backend.temp_schema}.{temp_table_name} where"
                        f" {filter_expr}",
                        partitions,
                    )
                )

    def _get_save_partitions(self, target_table, temp_table_name):
        partitions_to_save = [target_table.partitions] if target_table.partitions else []
        if target_table.has_dynamic_partition():
            partition_values = self.backend.exec_sql(
                f'select distinct {", ".join([p.field for p in target_table.partitions])} '
                f"from {self.backend.temp_schema}.{temp_table_name}"
            ).collect()
            partitions_to_save = [
                [Partition(p.field, v[i]) for i, p in enumerate(target_table.partitions)] for v in partition_values
            ]
        return partitions_to_save


class RdbRow(Row):
    def __init__(self, cols: List[str], values: Optional[Tuple]):
        self._cols = cols
        from decimal import Decimal

        # case decimal to float in order for later comparing (to ensure type consistency)
        self._values = tuple([float(v) if isinstance(v, Decimal) else v for v in values])

    def as_dict(self):
        return None if self._values is None else dict(zip(self._cols, self._values))

    def as_tuple(self):
        return self._values

    def __eq__(self, other: RdbRow):
        if (
            not isinstance(
                other,
                (
                    RdbRow,
                    tuple,
                ),
            )
            or other is None
        ):
            return False
        if isinstance(other, RdbRow):
            return other._cols == self._cols and other._values == self._values
        elif isinstance(other, tuple):
            return other == self._values

    def __str__(self):
        return f'({", ".join([f"{k}={_quote_str(v)}" for k, v in zip(self._cols, self._values)])})'

    def __getitem__(self, i):
        return self._values[i]

    def __repr__(self):
        return "RdbRow" + str(self)


class RdbBackend(Backend):
    """table_partitions_table_name;
    means the table name which save the static partition info for all partition tables in data warehouse,
    for now need support backend type: [clickhouse]
    others backend has another method to manage static partition info or just support static partition"""

    def __init__(
        self,
        url: str,
        credentials: str = None,
        sql_expr: SqlExpr = None,
        partitions_table_name="dataplat.__table_partitions__",
    ):
        self.partitions_table_name = partitions_table_name
        self.url, self.credentials = url, credentials
        self.sql_expr = sql_expr or SqlExpr()
        self.__init_inner(self.url, self.credentials)

    def __init_inner(self, url: str, credentials: str = None):
        from sqlalchemy import create_engine

        self.temp_schema = f"sp_temp_{int(time.mktime(time.gmtime()))}_{int(random() * 10000):04d}"

        self.backend_type, self.is_pg, self.is_ch, self.is_bq = None, False, False, False
        self.sql_dialect: SqlDialect = None
        if url.startswith("postgresql://"):
            self.backend_type, self.is_pg = "pg", True
            self.sql_dialect = PgSqlDialect(self.sql_expr)
            self.engine: Engine = create_engine(url, isolation_level="AUTOCOMMIT", pool_size=1)
            self.conn: Connection = self.engine.connect()
            _exec_sql(self.conn, self.sql_dialect.create_db_sql(self.temp_schema))
            _exec_sql(self.conn, self.sql_dialect.use_db_sql(self.temp_schema))
        elif url.startswith("clickhouse://") or url.startswith("clickhouse+native://"):
            self.backend_type, self.is_ch = "ch", True
            self.sql_dialect = ChSqlDialect(self.sql_expr, self.partitions_table_name)

            engine: Engine = create_engine(url, pool_size=1)
            conn: Connection = engine.connect()
            _exec_sql(conn, self.sql_dialect.create_db_sql(self.temp_schema))

            self._create_partitions_table(conn)

            conn.close()

            url_parts = url.split("?")
            url_params = "" if len(url_parts) == 1 else f"?{url_parts[1]}"
            url_raw_parts = url_parts[0].split("/")
            if len(url_raw_parts) == 4:  # db in url
                url_raw_parts = url_raw_parts[:-1]
            elif len(url_raw_parts) == 3:  # db not in url
                pass
            else:
                raise Exception(f"unrecognized url: {url}")
            url = f'{"/".join(url_raw_parts + [self.temp_schema])}{url_params}'

            self.engine: Engine = create_engine(url, pool_size=1)
            self.conn: Connection = self.engine.connect()
        elif url.startswith("bigquery://"):
            self.backend_type, self.is_bq = "bq", True
            self.sql_dialect = BqSqlDialect(self.temp_schema, self.sql_expr)
            self.engine: Engine = create_engine(url, credentials_path=credentials)
            self.conn: Connection = self.engine.connect()
            _exec_sql(self.conn, self.sql_dialect.create_db_sql(self.temp_schema))

    def _create_partitions_table(self, conn):
        cols = [
            {"name": "db_name", "type": "String"},
            {"name": "table_name", "type": "String"},
            {"name": "partition_value", "type": "String"},
            {"name": "last_modified_time", "type": "DateTime"},
        ]
        partitions = [Partition(field="db_name")]
        db_name = self.partitions_table_name.split(".")[0]
        _exec_sql(conn, self.sql_dialect.create_db_sql(db_name))
        _exec_sql(conn, self.sql_dialect.create_table_with_partitions_sql(self.partitions_table_name, cols, partitions))

    @property
    def inspector(self):
        from sqlalchemy import inspect

        # inspector object has cache built-in, so we should recreate the object if required
        inspector: Inspector = inspect(self.engine)
        return inspector

    def get_columns(self, table_name, schema=None, raw=False, **kw) -> List[Dict]:
        cols = self.inspector.get_columns(table_name, schema, **kw)
        if not raw:
            for col in cols:
                col_type: TypeEngine = col["type"]
                col["type"] = col_type.compile(self.inspector.dialect)
        return cols

    def get_column_names(self, table_name, schema=None, **kw) -> List[str]:
        cols = self.get_columns(table_name, schema, raw=True, **kw)
        names = [col["name"] for col in cols if "name" in col]
        return names

    def reset(self):
        if self.conn:
            try:  # noqa: SIM105
                self.conn.close()
            except Exception:
                pass
        if self.engine:
            try:  # noqa: SIM105
                self.engine.dispose()
            except Exception:
                pass
        self.__init_inner(self.url, self.credentials)

    def init_udfs(self, *args, **kwargs):
        self.register_udfs(udfs.get_udfs(self.backend_type))

    def register_udfs(self, funcs: Dict[str, Callable[[], Union[str, List[str]]]]):
        for udf_sql_creator in funcs.values():
            sql = udf_sql_creator()
            sqls = sql if isinstance(sql, list) else [sql]
            for sql in sqls:
                _exec_sql(self.conn, sql)

    def create_empty_table(self):
        return RdbTable(self, "")

    def exec_native_sql(self, sql: str) -> Any:
        return _exec_sql(self.conn, sql)

    def exec_sql(self, sql: str) -> RdbTable:
        return RdbTable(self, sql)

    def _tables(self, db: str) -> List[str]:
        all_tables = _exec_sql(self.conn, self.sql_dialect.get_tables_sql(db)).fetchall()
        return sorted([table[0] for table in all_tables])

    def _dbs(self) -> List[str]:
        all_schemas = _exec_sql(self.conn, self.sql_dialect.get_dbs_sql()).fetchall()
        return sorted([schema[0] for schema in all_schemas])

    def temp_tables(self) -> List[str]:
        return self._tables(self.temp_schema)

    def clear_cache(self):
        pass

    def clean(self):
        logger.info(f"clean temp db: {self.temp_schema}")
        _exec_sql(self.conn, self.sql_dialect.drop_db_sql(self.temp_schema))

    def clear_temp_tables(self, exclude: List[str] = None):
        from sqlalchemy.exc import ProgrammingError

        for table in self.temp_tables():
            if table not in exclude:
                print(f"dropping temp table {table}")
                try:
                    _exec_sql(self.conn, self.sql_dialect.drop_view_sql(table))
                except ProgrammingError as e:
                    if re.match(r'.*view ".*" does not exist', e.args[0]):
                        # Since we will drop view cascade in pg, so some view might already be dropped.
                        # It will raise the view-not-exist error, we just ignore this kind of error.
                        pass
                    else:
                        raise e

    def create_temp_table(self, table: RdbTable, name: str):
        logger.info(f"create_temp_table with: table={table}, name={name}")
        table.save_to_temp_table(name)

    def create_cache_table(self, table: RdbTable, name: str):
        logger.info(f"create_cache_table with: table={table}, name={name}")
        table.save_to_temp_table(name)

    def broadcast_table(self, table: RdbTable, name: str):
        logger.info(f"broadcast_table with: table={table}, name={name}")
        table.save_to_temp_table(name)

    def db_exists(self, db_name: str) -> bool:
        return db_name in self._dbs()

    def table_exists(self, table: TableMeta):
        db_name, table_name = table.dbname, table.pure_table_name
        db_name = db_name or self.temp_schema
        return self.db_exists(db_name) and table_name in self._tables(db_name)

    def refresh_table_partitions(self, table: TableMeta):
        if self.sql_dialect.support_native_partition():
            native_partitions_sql, extract_partition_cols = self.sql_dialect.native_partitions_sql(table.table_name)
            pt_cols = extract_partition_cols(_exec_sql(self.conn, native_partitions_sql))
            table.update_partitions([Partition(col) for col in pt_cols])
        # no need to do anything, if the db does not support partition

    def _get_save_partitions(self, original_source_table, source_table, target_table):
        # if original_source_table has dynamic partitions , it will generate multi partitions
        # if original_source_table has static partitions, it will generate only one partition which is target_table.partitions
        save_partitions_list = []
        if original_source_table.has_dynamic_partition():  # dynamic partitions (partition retrieved from real table)
            pt_cols = [pt.field for pt in original_source_table.partitions]
            pt_values_list = _exec_sql(
                self.conn,
                f'select distinct {", ".join(pt_cols)} from {source_table.get_full_table_name(self.temp_schema)}',
            ).fetchall()
            for pt_values in pt_values_list:
                save_partitions_list.append([Partition(field, value) for field, value in zip(pt_cols, pt_values)])
        else:  # static partitions (partition specified in sql file)
            save_partitions_list.append(target_table.partitions)
        return save_partitions_list

    def save_table(
        self, source_table: TableMeta, target_table: TableMeta, save_mode: SaveMode, create_target_table: bool
    ):
        logger.info(
            f"save table with: source_table={source_table}, target_table={target_table}, "
            f"save_mode={save_mode}, create_target_table={create_target_table}"
        )

        if not self.sql_dialect.support_static_partition():
            if not self.db_exists(target_table.dbname) and create_target_table:
                _exec_sql(self.conn, self.sql_dialect.create_db_sql(target_table.dbname))
            _exec_sql(self.conn, self.sql_dialect.create_pt_meta_table_sql(target_table.dbname))

        if not self.table_exists(target_table) and not create_target_table:
            raise Exception(
                f"target table {target_table.table_name} does not exist, and create_target_table is False, "
                f"cannot save table {source_table.table_name} to {target_table.table_name}"
            )

        source_table = source_table.clone_with_partitions(target_table.partitions)
        if not self.table_exists(target_table) and create_target_table:
            RdbTable.from_table_meta(self, source_table).save_to_table(target_table)
            return

        target_col_names = self.get_column_names(target_table.pure_table_name, target_table.dbname or self.temp_schema)
        original_source_table = source_table
        source_table = TableMeta(RdbTable.from_table_meta(self, source_table).resolve_to_temp_table())
        source_col_names = self.get_column_names(source_table.pure_table_name, source_table.dbname or self.temp_schema)
        logger.info(
            f"ensure cols match for source_table {source_table.table_name} and target_table {target_table.table_name}"
        )
        self._ensure_contain_target_cols(source_col_names, target_col_names)

        if save_mode == SaveMode.overwrite:
            self._overwrite(source_table, target_table, original_source_table, target_col_names)
        elif save_mode == SaveMode.append:
            self._append(source_table, target_table, original_source_table, target_col_names)
        else:
            raise SqlProcessorAssertionError(f"unknown save mode {save_mode}")

    def _ensure_contain_target_cols(self, source_cols: List[str], target_cols: List[str]):
        if not set(target_cols).issubset(set(source_cols)):
            raise Exception(
                f"source_cols does not contain target_cols: source_cols={source_cols}, target_cols={target_cols}"
            )

    def create_table_with_data(
        self, full_table_name: str, values: List[List[Any]], schema: List[Col], partitions: List[Partition]
    ):
        db, _ = full_table_name[: full_table_name.index(".")], full_table_name[full_table_name.index(".") + 1 :]
        _exec_sql(self.conn, self.sql_dialect.create_db_sql(db))
        _exec_sql(
            self.conn,
            self.sql_dialect.create_table_with_partitions_sql(
                full_table_name, [col.as_dict() for col in schema], partitions
            ),
        )
        cols = [col.name for col in schema]
        pt_cols = [p.field for p in partitions]
        pt_values_list = [[row[cols.index(p)] for p in pt_cols] for row in values]
        partitions = {
            tuple([Partition(field, value) for field, value in zip(pt_cols, pt_values)]) for pt_values in pt_values_list
        }
        if partitions and not self.sql_dialect.create_partition_automatically():
            for partition in partitions:
                if partition:
                    _exec_sql(self.conn, self.sql_dialect.create_partition_sql(full_table_name, list(partition)))
        from sqlalchemy.sql import text

        converted_col_names = ", ".join(self.sql_dialect.convert_pt_col_expr([f":{col}" for col in cols], pt_cols))
        stmt = text(f'insert into {full_table_name} ({", ".join(cols)}) VALUES ({converted_col_names})')
        for v in values:
            _exec_sql(self.conn, stmt, **{col: _v for _v, col in zip(v, cols)})

        if partitions and not self.sql_dialect.support_static_partition():
            _exec_sql(self.conn, self.sql_dialect.create_pt_meta_table_sql(db))
            for partition in partitions:
                if partition:
                    _exec_sql(self.conn, self.sql_dialect.insert_pt_metadata_sql(full_table_name, list(partition)))

    def create_temp_table_with_data(self, table_name: str, values: List[List[Any]], schema: List[Col]):
        _exec_sql(
            self.conn,
            self.sql_dialect.create_table_with_partitions_sql(table_name, [col.as_dict() for col in schema], []),
        )
        cols = [col.name for col in schema]
        from sqlalchemy.sql import text

        stmt = text(f'insert into {table_name} ({", ".join(cols)}) VALUES ({", ".join([f":{col}" for col in cols])})')
        for v in values:
            _exec_sql(self.conn, stmt, **{col: _v for _v, col in zip(v, cols)})

    def _overwrite(
        self,
        source_table: TableMeta,
        target_table: TableMeta,
        original_source_table: TableMeta,
        target_cols: List[str],
    ):
        col_names = ", ".join(target_cols)
        full_target_table_name = target_table.get_full_table_name(self.temp_schema)
        # write data to temp table to support the case when read from and write to the same table
        temp_table_name = f"{full_target_table_name}__temp"
        _exec_sql(self.conn, self.sql_dialect.drop_table_sql(temp_table_name))
        if self.table_exists(target_table):
            _exec_sql(
                self.conn,
                self.sql_dialect.create_table_like_sql(
                    temp_table_name, target_table.table_name, target_table.partitions
                ),
            )
        RdbTable.from_table_meta(self, source_table).save_to_table(target_table.clone_with_name(temp_table_name))
        if original_source_table.has_partitions():
            save_partitions = self._get_save_partitions(original_source_table, source_table, target_table)
            for save_partition in save_partitions:
                _exec_sql(self.conn, self.sql_dialect.delete_partition_sql(target_table.table_name, save_partition))
                if not self.sql_dialect.create_partition_automatically():
                    _exec_sql(self.conn, self.sql_dialect.create_partition_sql(full_target_table_name, save_partition))

                if self.sql_dialect.support_move_individual_partition():
                    _exec_sql(
                        self.conn,
                        self.sql_dialect.move_data_sql(full_target_table_name, temp_table_name, save_partition),
                    )
                else:
                    filter_expr = " and ".join(
                        [f"{pt.field} = {self.sql_expr.for_value(pt.value)}" for pt in save_partition]
                    )
                    _exec_sql(
                        self.conn,
                        self.sql_dialect.insert_data_sql(
                            full_target_table_name,
                            col_names,
                            f"select {col_names} from {temp_table_name} where {filter_expr}",
                            save_partition,
                        ),
                    )
            _exec_sql(self.conn, self.sql_dialect.drop_table_sql(temp_table_name))
        else:
            _exec_sql(self.conn, self.sql_dialect.drop_table_sql(full_target_table_name))
            _exec_sql(self.conn, self.sql_dialect.rename_table_sql(temp_table_name, full_target_table_name))

    def _append(
        self,
        source_table: TableMeta,
        target_table: TableMeta,
        original_source_table: TableMeta,
        target_cols: List[str],
    ) -> None:
        col_names = ", ".join(target_cols)
        full_target_table_name = target_table.get_full_table_name(self.temp_schema)
        if original_source_table.has_partitions():
            save_partitions = self._get_save_partitions(original_source_table, source_table, target_table)
            for save_partition in save_partitions:
                if not self.sql_dialect.create_partition_automatically():
                    _exec_sql(
                        self.conn,
                        self.sql_dialect.create_partition_sql(full_target_table_name, save_partition, True),
                    )
                filter_expr = " and ".join(
                    [f"{pt.field} = {self.sql_expr.for_value(pt.value)}" for pt in save_partition]
                )
                _exec_sql(
                    self.conn,
                    self.sql_dialect.insert_data_sql(
                        full_target_table_name,
                        col_names,
                        f"select {col_names} from {source_table.get_full_table_name(self.temp_schema)} where"
                        f" {filter_expr}",
                        save_partition,
                    ),
                )
        else:
            _exec_sql(
                self.conn,
                self.sql_dialect.insert_data_sql(
                    full_target_table_name,
                    col_names,
                    f"select {col_names} from {source_table.get_full_table_name(self.temp_schema)}",
                    [],
                ),
            )



from typing import Any, Callable, Dict, List, Union


def _all_udfs(cls: Any):
    return {
        attr: getattr(cls, attr)
        for attr in dir(cls)
        if callable(getattr(cls, attr)) and not attr.startswith("_") and attr != "all"
    }


def get_udfs(type: str) -> Union[Dict[str, Callable[[], Union[str, List[str]]]], Dict[str, Callable]]:
    if type == "pg":
        return PgUdfs.all()
    elif type == "ch":
        return ChUdfs.all()
    elif type == "spark":
        return SparkUdfs.all()
    else:
        return {}


class SparkUdfs:
    @staticmethod
    def all() -> Dict[str, Callable]:
        return _all_udfs(SparkUdfs)

    @staticmethod
    def remove_all_whitespaces(value: str) -> str:
        return "".join(value.split()) if value is not None else None

    @staticmethod
    def trim_all(value: str) -> str:
        return value.strip() if value is not None else None


class PgUdfs:
    @staticmethod
    def all() -> Dict[str, Callable[[], str]]:
        return _all_udfs(PgUdfs)

    @staticmethod
    def trim_all():
        return """
create or replace function trim_all(value text) returns text
    as $$ select regexp_replace(regexp_replace($1, E'^[\\\\a\\\\b\\\\e\\\\f\\\\n\\\\r\\\\t\\\\v\\\\0 ]+', ''), E'[\\\\a\\\\b\\\\e\\\\f\\\\n\\\\r\\\\t\\\\v\\\\0 ]+$', '') $$
    LANGUAGE SQL
    IMMUTABLE
    RETURNS NULL ON NULL INPUT
"""  # noqa : B950

    @staticmethod
    def split():
        return """
create or replace function split(value text, sep text) returns text[]
    as $$ select string_to_array($1, $2) $$
    LANGUAGE SQL
    IMMUTABLE
    RETURNS NULL ON NULL INPUT
"""

    @staticmethod
    def from_unixtime():
        return """
create or replace function from_unixtime(value float) returns timestamp
    as $$ select to_timestamp($1) $$
    LANGUAGE SQL
    IMMUTABLE
    RETURNS NULL ON NULL INPUT
"""

    @staticmethod
    def date_format():
        return """
create or replace function date_format(value timestamp, format text) returns text
    as $$ select to_char($1, $2) $$
    LANGUAGE SQL
    IMMUTABLE
    RETURNS NULL ON NULL INPUT
"""

    @staticmethod
    def get_json_object():
        return """
create or replace function get_json_object(value text, path text) returns text
    as $$ select $1::json#>(string_to_array($2, '.'))[2:] $$
    LANGUAGE SQL
    IMMUTABLE
    RETURNS NULL ON NULL INPUT
"""

    @staticmethod
    def sha1():
        return """
CREATE EXTENSION IF NOT EXISTS pgcrypto with schema public;
create or replace function sha1(value text) returns text
    as $$ select encode(public.digest($1::bytea, cast('sha1' as text)), 'hex') $$
    LANGUAGE SQL
    IMMUTABLE
    RETURNS NULL ON NULL INPUT
""".split(
            ";"
        )


class ChUdfs:
    """
    https://clickhouse.com/docs/en/sql-reference/statements/create/function
    CREATE FUNCTION name AS (parameter0, ...) -> expression
    """

    @staticmethod
    def all() -> Dict[str, Callable[[], str]]:
        return _all_udfs(ChUdfs)

    @staticmethod
    def translate():
        return """
CREATE FUNCTION IF NOT EXISTS translate AS (input, from, to) -> replaceAll(input, from, to)
        """



from __future__ import annotations

import json
import os
import re
import sys
import unittest
from datetime import datetime
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union

import openpyxl
from bson import json_util

__all__ = ["SqlTester", "TableData", "TestCase", "WorkPath", "work_path", "SqlReader", "TableColumnTypes"]

from easy_sql.logger import logger
from easy_sql.sql_processor import SqlProcessor
from easy_sql.sql_processor.backend import Backend, Partition, Row, SparkBackend
from easy_sql.sql_processor.backend.rdb import Col, RdbBackend

if TYPE_CHECKING:
    from openpyxl.cell import Cell
    from openpyxl.workbook import Workbook
    from openpyxl.worksheet.worksheet import Worksheet
    from pyspark.sql import SparkSession
    from pyspark.sql.types import StructType

debug_log_enable = True


def log_debug(msg: str):
    if debug_log_enable:
        logger.debug(msg)


class TableData:
    def __init__(
        self,
        name: str,
        columns: List[str],
        column_types: List[str],
        values: List[List[Any]],
        value_descriptions: List[str],
    ):
        self.name, self.columns, self.column_types, self.values, self.value_descriptions = (
            name,
            columns,
            column_types,
            values,
            value_descriptions,
        )

    def as_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "columns": json.dumps(self.columns),
            "column_types": json.dumps(self.column_types),
            "value_descriptions": self.value_descriptions,
            "values": [json.dumps(v, default=json_util.default, ensure_ascii=False) for v in self.values],
        }

    def pt_col(self, possible_pt_cols: List[str]) -> Optional[str]:
        for col in possible_pt_cols:
            if col in self.columns:
                return col
        return None

    @staticmethod
    def from_dict(data: Dict) -> TableData:
        return TableData(
            data["name"],
            json.loads(data["columns"]),
            json.loads(data["column_types"]),
            [json.loads(v, object_hook=json_util.object_hook) for v in data["values"]],
            data["value_descriptions"],
        )


class lazy_property:
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, cls):
        val = self.func(instance)
        setattr(instance, self.func.__name__, val)
        return val


class WorkPath:
    def __init__(self, work_path: str = None):
        self._work_path: Optional[str] = work_path

    def set_work_path(self, work_path: str):
        self._work_path = work_path

    def work_path(self):
        if not self._work_path:
            raise AssertionError("work_path must be set before use.")
        return self._work_path

    def path(self, relative_path: str):
        return os.path.join(self.work_path(), relative_path)

    def relative_path(self, abs_path: str):
        if not abs_path.startswith(self.work_path()):
            raise AssertionError(f"abs path `{abs_path}` not in work directory: ")
        return "" if len(abs_path) == self.work_path() else abs_path[len(self.work_path()) + 1 :]


work_path = WorkPath()


class TableColumnTypes:
    def __init__(
        self, predefined_table_col_types: Dict[str, Dict[str, str]], partition_col_types: Dict[str, str], backend: str
    ):
        self.table_col_types = predefined_table_col_types
        assert set(partition_col_types.values()).issubset(
            {"int", "string", "Int32", "String"}
        ), f"only int/string partition column types supported, found {partition_col_types}"
        self.partition_col_types = partition_col_types
        self.backend = backend

    def get_col_type(self, table_name: str, col_name: str) -> str:
        col_name = col_name.lower()
        if table_name not in self.table_col_types:
            raise Exception(f"table definition not found from any table files: {table_name}")
        if col_name not in self.table_col_types[table_name]:
            if col_name in self.partition_col_types:
                return self.partition_col_types[col_name]
            raise Exception(
                f"column {col_name} not found for table {table_name}, known columns::"
                f" {list(self.table_col_types[table_name].keys())}"
            )
        if self.backend == "clickhouse":
            return self.table_col_types[table_name][col_name].strip()
        return self.table_col_types[table_name][col_name].lower().strip()

    def column_types_to_schema(
        self, backend: Backend, columns: List[str], types: List[str]
    ) -> Union[StructType, List[Col]]:
        if isinstance(backend, SparkBackend):
            return self.column_types_to_schema_spark(backend.spark, columns, types)
        elif isinstance(backend, RdbBackend):
            return self.column_types_to_schema_rdb(backend, columns, types)
        else:
            raise AssertionError(f"unsupported backend type: {type(backend)}")

    def column_types_to_schema_rdb(self, backend: RdbBackend, columns: List[str], types: List[str]) -> List[Col]:
        return [Col(col, type_) for col, type_ in zip(columns, types)]

    def column_types_to_schema_spark(self, spark: SparkSession, columns: List[str], types: List[str]) -> StructType:
        from pyspark.sql.functions import expr
        from pyspark.sql.types import (
            ArrayType,
            BooleanType,
            DateType,
            DoubleType,
            FloatType,
            IntegerType,
            LongType,
            ShortType,
            StringType,
            StructField,
            StructType,
            TimestampType,
        )

        fields_creators = {
            "int": lambda name: StructField(name, IntegerType()),
            "tinyint": lambda name: StructField(name, ShortType()),
            "bigint": lambda name: StructField(name, LongType()),
            "double": lambda name: StructField(name, DoubleType()),
            "float": lambda name: StructField(name, FloatType()),
            "string": lambda name: StructField(name, StringType()),
            "decimal": lambda name: StructField(name, DoubleType()),
            "boolean": lambda name: StructField(name, BooleanType()),
            "date": lambda name: StructField(name, DateType()),
            "timestamp": lambda name: StructField(name, TimestampType()),
            "array<string>": lambda name: StructField(name, ArrayType(StringType())),
            "array<int>": lambda name: StructField(name, ArrayType(IntegerType())),
            "array<tinyint>": lambda name: StructField(name, ArrayType(ShortType())),
            "array<bigint>": lambda name: StructField(name, ArrayType(LongType())),
            "array<double>": lambda name: StructField(name, ArrayType(DoubleType())),
            "array<float>": lambda name: StructField(name, ArrayType(FloatType())),
            "array<boolean>": lambda name: StructField(name, ArrayType(BooleanType())),
            "array<date>": lambda name: StructField(name, ArrayType(DateType())),
            "array<timestamp>": lambda name: StructField(name, ArrayType(TimestampType())),
        }

        fields = []
        for col, type_ in zip(columns, types):
            type_ = "decimal" if type_.startswith("decimal(") else type_
            if type_ in fields_creators:
                fields.append(fields_creators[type_](col))
            else:
                logger.info(f"try to create null value for `{type_}` when convert to spark schema for column `{col}`")
                field_schema = (
                    spark.createDataFrame(spark.sparkContext.emptyRDD(), StructType([]))
                    .withColumn(col, expr(f"cast(null as {type_})"))
                    .schema.fields[0]
                )
                fields.append(field_schema)

        return StructType(fields)

    def cast_as_type(
        self, table_name: str, col_name: str, col_value: Any, date_converter: Callable = None, col_type: str = None
    ) -> Tuple[str, Any]:
        col_type = self.get_col_type(table_name, col_name) if col_type is None else col_type
        if self.backend == "clickhouse":
            if "Nested" in col_type:
                raise AssertionError("clickhouse backend can not support Nested col type")
            # for these type in clickhouse is case insensitive
            if col_type.lower() in ["bool", "date", "datetime", "decimal"]:
                col_type = col_type.lower()
        else:
            col_type = col_type.lower()
        if col_value is None or (isinstance(col_value, str) and col_value.strip() == "null"):
            return col_type, None

        def process_pt_val(date_converter: Callable, col_value: str):
            if not date_converter:
                return str(col_value).strip()
            return date_converter(col_value).strftime("%Y-%m-%d") if date_converter(col_value) else None

        try:
            if col_name in self.partition_col_types:
                backend_is_bigquery = self.backend == "bigquery"
                if backend_is_bigquery or col_name == "data_date":
                    # for bigquery, partition column must be a date
                    return col_type, process_pt_val(date_converter, col_value)
                else:
                    try:
                        # try convert partition column to int to remove possible float values
                        return col_type, str(round(float(col_value))) if col_type == "string" else round(
                            float(col_value)
                        )
                    except ValueError as e:
                        if col_type == "string":
                            return col_type, str(col_value)
                        else:
                            raise e

            if col_type.replace(" ", "").startswith("map<"):
                raise AssertionError(
                    f"map type not supported right now when parsing value `{col_value}` for column:"
                    f" {table_name}.{col_name}"
                )
            if col_type.startswith("decimal(") or col_type == "double":
                return col_type, float(col_value)
            if col_type in ["bigint", "int", "tinyint", "Int32", "Int64", "UInt32", "UInt64"]:
                return col_type, int(col_value)
            if col_type in ["boolean", "bool"]:
                if str(col_value).lower() == "true":
                    return col_type, True
                elif str(col_value).lower() == "false":
                    return col_type, False
                return col_type, bool(col_value)
            if col_type in ["string", "text", "String"]:
                return col_type, str(col_value).strip()
            if col_type.replace(" ", "") in ["array<string>", "text[]", "Array(String)"]:
                return col_type, [s.strip() for s in str(col_value).strip().split("|") if s.strip()]
            if col_type.replace(" ", "") in ["array<int>", "int[]", "Array(Int)"]:
                return col_type, [int(s.strip()) for s in str(col_value).strip().split("|") if s.strip()]
            if col_type.replace(" ", "") == "struct<latest_value:string,first_show_time:timestamp>":
                vals = str(col_value).strip().split("|")
                if len(vals) < 2:
                    raise AssertionError(
                        f"must provide all the values of type {col_type} for {table_name}.{col_name}. Incomplete value"
                        f" `{col_value}`"
                    )
                latest_value = str(vals[0]).strip()
                first_show_time = str(vals[1]).strip() if not date_converter else date_converter(vals[1])
                return col_type, (latest_value, first_show_time)
            if col_type in ["date", "timestamp", "datetime"]:
                if not date_converter:
                    return col_type, str(col_value).strip()
                else:
                    return col_type, date_converter(col_value)
        except Exception as e:
            raise Exception(f"convert {col_type} failed for {table_name}.{col_name} for value `{col_value}`", e)

        raise AssertionError(
            f"not supported type {col_type} for column {table_name}.{col_name} when parsing value {col_value}"
        )


class TestCase:
    def __init__(self, sql_file_path: str = None, sql_file_content: str = None, default_col_type="string"):
        self.name, self.vars = None, {}
        self.includes = {}
        self.inputs: List[TableData] = []
        self.outputs: List[TableData] = []
        assert sql_file_path is not None or sql_file_content is not None, "sql_file_path or sql_file_content required"
        self.sql_file_path = sql_file_path
        self.sql_file_content = sql_file_content
        self.udf_file_paths: List[str] = []
        self.func_file_paths: List[str] = []
        self.default_col_type = default_col_type

    def as_dict(self):
        data = {
            attr: getattr(self, attr)
            for attr in dir(self)
            if not attr.startswith("_") and not callable(getattr(self, attr))
        }
        data.update(
            {
                "inputs": [table_data.as_dict() for table_data in self.inputs],
                "outputs": [table_data.as_dict() for table_data in self.outputs],
            }
        )
        return data

    @staticmethod
    def from_dict(data: Dict) -> TestCase:
        case = TestCase(data["sql_file_path"], data.get("sql_file_content", None))
        case.name = data["name"]
        case.vars = data["vars"]
        case.includes = data["includes"] if "includes" in data else {}
        case.inputs = [TableData.from_dict(table_data_dict) for table_data_dict in data["inputs"]]
        case.outputs = [TableData.from_dict(table_data_dict) for table_data_dict in data["outputs"]]
        case.udf_file_paths = data["udf_file_paths"] if "udf_file_paths" in data else []
        case.func_file_paths = data["func_file_paths"] if "func_file_paths" in data else []
        return case

    @property
    def simple_sql_name(self):
        return os.path.basename(self.sql_file_path) if self.sql_file_path else None

    def parse_test_case_of_label(
        self, wb: Workbook, label: str, row_start_idx: int, rows: List[List[Cell]], table_column_types: TableColumnTypes
    ):
        log_debug(f"start to parse case from row {row_start_idx} for label {label}")
        if label == "CASE":
            self.name = str(rows[0][1].value).strip()
        elif label == "VARS":
            self.parse_vars(wb, row_start_idx, rows)
        elif label == "INCLUDES":
            self.parse_includes(wb, row_start_idx, rows)
        elif label == "INPUT":
            self.parse_input(wb, row_start_idx, rows, table_column_types)
        elif label == "OUTPUT":
            self.parse_output(wb, row_start_idx, rows, table_column_types)
        elif label == "UDFS":
            self.parse_udfs(row_start_idx, rows)
        elif label == "FUNCS":
            self.parse_funcs(row_start_idx, rows)
        else:
            raise AssertionError(f"unknown label: {label}")

    def parse_includes(self, wb: Workbook, row_start_idx: int, rows: List[List[Cell]]):
        for i, row in enumerate(rows):
            include_name, include_value = row[1].value and row[1].value.strip(), row[2].value and row[2].value.strip()
            if include_name:
                if not include_value:
                    raise AssertionError(
                        f"parse test case at row {row_start_idx + 1 + i} failed: there must be value set for INCLUDES,"
                        " found None"
                    )
                self.includes[include_name] = include_value
                log_debug(
                    f"find includes at row {row_start_idx + 1}: include_name={include_name},"
                    f" include_value={include_value}"
                )

    def parse_udfs(self, row_start_idx: int, rows: List[List[Cell]]):
        for udf_path in rows[0][1:]:
            if udf_path.value and udf_path.value.strip():
                self.udf_file_paths.append(udf_path.value)
                log_debug(f"find udf_path at row {row_start_idx + 1}: udf_relative_path={udf_path.value}")

    def parse_funcs(self, row_start_idx: int, rows: List[List[Cell]]):
        for func_path in rows[0][1:]:
            if func_path.value and func_path.value.strip():
                self.func_file_paths.append(func_path.value)
                log_debug(f"find func_path at row {row_start_idx + 1}: func_relative_path={func_path.value}")

    def parse_vars(self, wb: Workbook, row_start_idx: int, rows: List[List[Cell]]):
        if len(rows) < 2:
            raise AssertionError(
                f"parse test case at row {row_start_idx + 1} failed: there must be value set for VARS, found None"
            )
        else:
            for var_name, var_value in zip(rows[0][1:], rows[1][1:]):
                if var_name.value and var_name.value.strip():
                    var_name, var_value = self.parse_var_from_cell(wb, var_name, var_value)
                    self.vars[var_name] = var_value
                    log_debug(
                        f"find vars at row {row_start_idx + 1}: name={var_name}, value={var_value} (type:"
                        f" {type(var_value)})"
                    )

    def parse_var_from_cell(self, wb: Workbook, var_name: Cell, var_value: Cell) -> Tuple[str, Any]:
        name = var_name.value.strip()
        if name.lower() == "data_date":
            value = self.parse_cell_value_as_date(wb, var_value.value).strftime("%Y-%m-%d")
        else:
            value = var_value.value
        return name, value

    def parse_cell_value_as_date(self, wb: Workbook, value: Any) -> datetime:
        if value is None or (isinstance(value, str) and value.strip() == ""):
            return None
        elif isinstance(value, str) and value.strip() != "":
            value = value.strip()
            if len(value) != len("2000-01-01") and len(value) != len("2000-01-01 00:00:00"):
                raise AssertionError("date column must be of format `yyyy-MM-dd` or `yyyy-MM-dd HH:mm:ss`")
            format = "%Y-%m-%d" if len(value) == len("2000-01-01") else "%Y-%m-%d %H:%M:%S"
            return datetime.strptime(value, format)
        return value

    def parse_output(
        self, wb: Workbook, row_start_idx: int, rows: List[List[Cell]], table_column_types: TableColumnTypes
    ):
        self.outputs.append(self.parse_table(wb, "OUTPUT", row_start_idx, rows, table_column_types))

    def parse_input(
        self, wb: Workbook, row_start_idx: int, rows: List[List[Cell]], table_column_types: TableColumnTypes
    ):
        self.inputs.append(self.parse_table(wb, "INPUT", row_start_idx, rows, table_column_types))

    def parse_table(
        self, wb: Workbook, label: str, row_start_idx: int, rows: List[List[Cell]], table_column_types: TableColumnTypes
    ) -> TableData:
        cells = rows[0]
        if len(cells) < 2 or not cells[1].value or not str(cells[1].value).strip():
            raise AssertionError(
                f"parse test case at row {row_start_idx + 1} failed: there must be table name set for {label}, found"
                " None"
            )
        if len(cells) < 3 or not cells[2].value or not str(cells[2].value).strip():
            raise AssertionError(
                f"parse test case at row {row_start_idx + 1} failed: there must be at least one column set for {label},"
                " found 0"
            )

        table_name = str(cells[1].value)

        columns, column_types = [], []
        type_in_column_names = self.is_type_in_column_names(cells)

        for cell in cells[2:]:
            if not cell.value or not str(cell.value).strip():
                break
            column_name = str(cell.value).strip()
            if type_in_column_names:
                if ":" in column_name:
                    columns.append(column_name[: column_name.index(":")])
                    column_types.append(column_name[column_name.index(":") + 1 :])
                else:
                    columns.append(column_name)
                    column_types.append(self.default_col_type)  # if no type specified, use string as default
            else:
                columns.append(column_name)
                if "." in table_name:
                    column_types.append(table_column_types.get_col_type(table_name, columns[-1]))

        values, value_descriptions = [], []
        for row_idx, cells in enumerate(rows[1:]):
            value_cells = cells[2:]
            has_values = any([value_cells[i].value not in [None, ""] for i in range(len(columns))])
            if cells[1].value and str(cells[1].value).strip():
                value_descriptions.append(str(cells[1].value).strip())
                values.append(
                    self._parse_table_row_values(
                        wb, table_name, columns, row_idx, row_start_idx, value_cells, column_types, table_column_types
                    )
                )
            elif label == "INPUT":
                # If no description mentioned for this row, just ignore it.
                # This ensures we must add description for a row, to clarify how the data comes.
                if has_values:
                    raise AssertionError(
                        f"no description for table({table_name}) data at row {row_start_idx + 1 + row_idx + 1}"
                    )
            elif label == "OUTPUT":
                has_values = any([value_cells[i].value not in [None, ""] for i in range(len(columns))])
                if has_values:
                    values.append(
                        self._parse_table_row_values(
                            wb,
                            table_name,
                            columns,
                            row_idx,
                            row_start_idx,
                            value_cells,
                            column_types,
                            table_column_types,
                        )
                    )

        log_debug(
            f'find {label.lower()} table at row {row_start_idx + 1}: {table_name}{{{", ".join(columns)}}}, data length'
            f" is {len(values)}"
        )
        print(f"table data for `{table_name}`: ")
        for v in values:
            print(v)
        return TableData(table_name, columns, column_types, values, value_descriptions)

    def is_type_in_column_names(self, cells) -> bool:
        for cell in cells[2:]:
            if not cell.value or not str(cell.value).strip():
                break
            column_name = str(cell.value).strip()
            if ":" in column_name:
                return True
        return False

    def _parse_table_row_values(
        self,
        wb: Workbook,
        table_name: str,
        columns: List[str],
        row_idx: int,
        row_start_idx: int,
        value_cells: List[Cell],
        column_types: List[str],
        table_column_types: TableColumnTypes,
    ):
        row_values = []
        for col_idx in range(len(columns)):
            try:
                _, value = table_column_types.cast_as_type(
                    table_name,
                    columns[col_idx],
                    value_cells[col_idx].value,
                    date_converter=lambda v: self.parse_cell_value_as_date(wb, v),
                    col_type=column_types[col_idx] if len(column_types) > 0 else None,
                )
                row_values.append(value)
            except Exception as e:
                if e.args[0].startswith("convert "):
                    raise Exception(
                        f"when parsing values at row: {row_start_idx + 1 + row_idx + 1}, col: {col_idx + 2 + 1}", e
                    )
                else:
                    raise e
        return row_values

    def read_sql_content(self) -> str:
        if self.sql_file_content:
            return self.sql_file_content
        else:
            if self.sql_file_path is None:
                raise AssertionError("can not find the sql file having same name with test file")
            with open(work_path.path(self.sql_file_path), "r") as f:
                return f.read()

    @property
    def completed(self):
        return self.name and len(self.inputs) and len(self.outputs)

    @property
    def missed_fields(self):
        fields = []
        if not self.name:
            fields.append("name")
        if not len(self.inputs):
            fields.append("inputs")
        if not len(self.outputs):
            fields.append("outputs")
        return fields


class TestDataFile:
    def __init__(self, test_data_file: str, sql_reader: SqlReader, backend: str = "spark"):
        self.test_data_file = test_data_file
        self.sql_reader = sql_reader
        self.wb = openpyxl.load_workbook(self.test_data_file)
        self.backend = backend

    def parse_test_cases(self, table_column_types: TableColumnTypes) -> List[TestCase]:
        wb = self.wb
        test_cases = []
        for sheet_name in wb.sheetnames:
            if not sheet_name.lower().startswith("suit"):
                continue
            for case in self.parse_test_cases_from_sheet(wb[sheet_name], table_column_types):
                test_cases.append(case)
        return test_cases

    def parse_test_cases_from_sheet(self, sheet: Worksheet, table_column_types: TableColumnTypes) -> List[TestCase]:
        cases = []
        last_row_idx = -1
        cases_rows = []
        cases_start_indices = []
        for row_idx, row in enumerate(sheet.rows):
            cells = list(row)
            if cells[0].value and cells[0].value.strip() == "CASE":
                if last_row_idx == -1:
                    last_row_idx = row_idx
                log_debug(f"add case at: {row_idx}")
                cases_start_indices.append(row_idx)
                cases_rows.append([cells])
            elif last_row_idx != -1:
                cases_rows[-1].append(cells)
        for case_start_idx, case_rows in zip(cases_start_indices, cases_rows):
            cases.append(self.parse_test_case(case_start_idx, case_rows, table_column_types))
        return cases

    def parse_test_case(
        self, case_start_idx: int, case_rows: List[List[Cell]], table_column_types: TableColumnTypes
    ) -> TestCase:
        if self.backend == "clickhouse":
            default_col_type = "String"
        else:
            default_col_type = "string"
        if self.sql_reader.read_as_content(self.test_data_file):
            case = TestCase(
                sql_file_content=self.sql_reader.read_sql(self.test_data_file), default_col_type=default_col_type
            )
        else:
            case = TestCase(
                self.sql_reader.find_file_path(self.test_data_file[: self.test_data_file.rindex(".")] + ".sql")
            )
        last_label, last_label_idx = None, -1
        for i, row in enumerate(case_rows):
            cells = row
            label = cells[0].value and cells[0].value.strip()
            if label in ["CASE", "VARS", "INCLUDES", "INPUT", "OUTPUT", "UDFS", "FUNCS"]:
                if last_label is not None:
                    case.parse_test_case_of_label(
                        self.wb,
                        last_label,
                        case_start_idx + last_label_idx,
                        case_rows[last_label_idx:i],
                        table_column_types,
                    )
                last_label, last_label_idx = label, i
        if last_label:
            case.parse_test_case_of_label(
                self.wb, last_label, case_start_idx + last_label_idx, case_rows[last_label_idx:], table_column_types
            )

        if not case.completed:
            raise AssertionError(f"parse test case failed, got incomplete case, missed fields: {case.missed_fields}")

        return case


class TestResult:
    PASSED = "PASSED"
    FAILED = "FAILED"

    def __init__(self, test_data_file: str):
        self.test_data_file = test_data_file
        self.case_results = []

    def collect_case_result(self, case_name: str, result: str):
        self.case_results.append({"case_name": case_name, "result": result})

    @property
    def failed_cases(self):
        return list(filter(lambda cr: cr["result"] == TestResult.FAILED, self.case_results))

    @property
    def is_fail(self) -> bool:
        return any([r["result"] == TestResult.FAILED for r in self.case_results])

    @property
    def is_success(self) -> bool:
        return not self.is_fail

    @property
    def passed_cases(self):
        return list(filter(lambda cr: cr["result"] == TestResult.PASSED, self.case_results))

    @property
    def simple_stat_str(self) -> str:
        return f"{len(self.passed_cases)} PASSED, {len(self.failed_cases)} FAILED"

    def print_result(self):
        print(f"test result for {self.test_data_file}: {self.simple_stat_str}")
        if len(self.failed_cases):
            print(f"failed cases: {', '.join([cr['case_name'] for cr in self.failed_cases])}")

    @staticmethod
    def print_results(test_results: List[TestResult]):
        print("======================Test Report====================")
        for tr in test_results:
            tr.print_result()
        failed_count = sum([len(tr.failed_cases) for tr in test_results])
        passed_count = sum([len(tr.passed_cases) for tr in test_results])
        if failed_count == 0:
            print(f"\nCongratulation! All {passed_count} tests passed!")
        else:
            print(f"\nThere are test failures! {passed_count} passed, {failed_count} failed")


class TestCaseRunner:
    def __init__(
        self,
        env: str,
        dry_run: bool,
        backend_creator: Callable,
        table_column_types: TableColumnTypes,
        unit_test_case: unittest.TestCase,
        sql_processor_creator: Callable,
    ):
        self.unit_test_case = unit_test_case or unittest.TestCase()
        self.env, self.dry_run, self.backend_creator = env, dry_run, backend_creator
        self.sql_processor_creator = sql_processor_creator
        self.table_column_types = table_column_types

    def run_test(self, case: TestCase):
        sql = case.read_sql_content()

        backend = self.backend_creator(case)
        try:
            self.clean(case, backend)
            self.create_inputs(case, backend)

            sql_processor = self.create_sql_processor(backend, case, sql)

            sql_processor.run(self.dry_run)

            self.verify_outputs(backend, case)
        finally:
            backend.clean()

    def verify_outputs(self, backend: Backend, case: TestCase):
        tempviews = backend.temp_tables()
        print("tempviews after test:", tempviews)
        for output in case.outputs:
            tempview_name = self.find_temp_view_for_output(case, output, tempviews)
            actual_output, expected_output = self.get_data(backend, output, tempview_name)

            def list_item_to_set(values: List[Union[List, Row]]):
                result = []
                for row in values:
                    result.append([set(v) if isinstance(v, list) else v for v in row])
                return result

            print("will verify equality for output: ", output.name)
            print("expected output: ", list_item_to_set(expected_output))
            print("actual output: ", list_item_to_set(actual_output))
            self.unit_test_case.assertListEqual(list_item_to_set(expected_output), list_item_to_set(actual_output))

    def create_sql_processor(self, backend: Backend, case: TestCase, sql: str) -> SqlProcessor:
        sql_processor = self.sql_processor_creator(backend, sql, case)
        for udf_file in case.udf_file_paths:
            sql_processor.register_udfs_from_pyfile(work_path.path(udf_file))
            log_debug(f"registering udf from `{work_path.path(udf_file)}` for case `{case.name}`")
        for func_file in case.func_file_paths:
            sql_processor.register_funcs_from_pyfile(work_path.path(func_file))
            log_debug(f"registering funcs from `{work_path.path(func_file)}` for case `{case.name}`")
        return sql_processor

    def get_data(self, backend: Backend, output: TableData, tempview_name: str) -> Tuple[List[Row], List[Row]]:
        full_tempview_name = f"{backend.temp_schema}.{tempview_name}" if backend.is_bigquery_backend else tempview_name
        select_output_sql = (
            f'select {", ".join(output.columns)} from {full_tempview_name} order by {", ".join(output.columns)}'
        )
        actual_output = backend.exec_sql(select_output_sql).collect()

        schema = self.table_column_types.column_types_to_schema(backend, output.columns, output.column_types)
        backend.create_temp_table_with_data(f"{full_tempview_name}__expected", output.values, schema)
        select_output_sql = (
            f'select {", ".join(output.columns)} from {full_tempview_name}__expected order by'
            f' {", ".join(output.columns)}'
        )
        expected_output = backend.exec_sql(select_output_sql).collect()
        return actual_output, expected_output

    def find_temp_view_for_output(self, case: TestCase, output: TableData, tempviews: List[str]) -> str:
        # tempview name format: {output_pure_table_name}_{md5(xxx)}
        hex_pattern = r"^[a-f0-9]+_output$" if self.dry_run else r"^[a-f0-9]+$"
        if "." in output.name:
            tempview_name = output.name[output.name.find(".") + 1 :]
            tempview_name = [
                tv
                for tv in tempviews
                if tv.startswith(tempview_name) and re.match(hex_pattern, tv[len(tempview_name) + 1 :])
            ]
        else:
            tempview_name = [tv for tv in tempviews if tv == output.name]
        if len(tempview_name) == 0:
            raise Exception(
                f"output `{output.name}` not found after execute test: {case.simple_sql_name}.{case.name}. "
                f"All temporary views are: {tempviews}"
            )
        elif len(tempview_name) > 1:
            raise Exception(
                f"multiple temp views found for output `{output.name}` found after execute test: {tempview_name}"
            )
        else:
            tempview_name = tempview_name[0]
        return tempview_name

    def create_inputs(self, case: TestCase, backend: Backend):
        for input in case.inputs:
            schema = self.table_column_types.column_types_to_schema(backend, input.columns, input.column_types)
            if "." in input.name:
                print(f"creating table: {input.name}")
                pt_col = input.pt_col(list(self.table_column_types.partition_col_types.keys()))
                backend.create_table_with_data(input.name, input.values, schema, [Partition(pt_col)] if pt_col else [])
            else:
                print(f"creating temp table: {input.name}")
                backend.create_temp_table_with_data(input.name, input.values, schema)

    def clean(self, case: TestCase, backend: Backend):
        databases = set()
        table_names = set.union({input.name for input in case.inputs}, {output.name for output in case.outputs})
        for table_name in table_names:
            if "." in table_name:
                databases.add(table_name.split(".")[0])
        for db in databases:
            try:
                if backend.is_bigquery_backend:
                    backend.exec_native_sql(f"drop schema if exists {db} cascade")
                elif backend.is_clickhouse_backend:
                    backend.exec_native_sql(f"drop database if exists {db}")
            except Exception as e:
                # BigQuery will throw an exception when deleting a nonexistent dataset even if using [IF EXISTS]
                import re

                if not re.match(
                    r"[\s\S]*Permission bigquery.datasets.delete denied on dataset[\s\S]*(or it may not exist)[\s\S]*",
                    str(e.args[0]),
                ):
                    raise e


class SqlTester:
    def __init__(
        self,
        backend_creator: Callable[[TestCase], Backend] = None,
        table_column_types: TableColumnTypes = None,
        sql_reader_creator: Callable[[], SqlReader] = None,
        sql_processor_creator: Callable[[Backend, str, TestCase], SqlProcessor] = None,
        unit_test_case: unittest.TestCase = None,
        dry_run: bool = True,
        env: str = "test",
        work_dir: str = None,
        backend: str = None,
        scala_udf_initializer: str = None,
    ):
        if work_dir is not None:
            work_path.set_work_path(os.path.abspath(work_dir))

        def create_sql_processor(backend, sql, case):
            if backend.is_bigquery_backend:
                vars = dict(case.vars.items() | {"temp_db": backend.temp_schema}.items())
            else:
                vars = case.vars
            return SqlProcessor(
                backend, sql, [], variables=vars, scala_udf_initializer=scala_udf_initializer, includes=case.includes
            )

        self.sql_processor_creator = sql_processor_creator or create_sql_processor
        self.backend = backend
        self.table_column_types = table_column_types or TableColumnTypes({}, {}, backend)
        self.unit_test_case = unit_test_case
        self.dry_run = dry_run
        self.env = env
        self.sql_reader = sql_reader_creator() if sql_reader_creator else SqlReader()
        self.test_case_runner = TestCaseRunner(
            self.env,
            self.dry_run,
            backend_creator,
            self.table_column_types,
            sql_processor_creator=self.sql_processor_creator,
            unit_test_case=self.unit_test_case,
        )

    def run_tests(self, test_data_files: List[str]):
        test_results = []
        for f in test_data_files:
            print(f"============= running test from file {os.path.basename(f)} [START] ==============")
            test_result = self.run_test(f)
            test_results.append(test_result)
            print(
                f"============= running test from file {os.path.basename(f)} [END:"
                f" {test_result.simple_stat_str}]=============="
            )
        TestResult.print_results(test_results)
        failed_count = sum([len(test_result.failed_cases) for test_result in test_results])
        if failed_count:
            sys.exit(1)

    def run_test(self, test_data_file: str, case_idx: int = -1) -> TestResult:
        cases = self.parse_test_cases(test_data_file, self.table_column_types)

        if case_idx != -1 and (case_idx < 0 or case_idx >= len(cases)):
            raise AssertionError(f"test case {case_idx} not found. {len(cases)} cases in {test_data_file} are found.")
        cases = cases if case_idx == -1 else [cases[case_idx]]

        tr = TestResult(test_data_file)
        for case in cases:
            passed = self.run_case(case)
            tr.collect_case_result(case.name, TestResult.PASSED if passed else TestResult.FAILED)
        return tr

    def parse_test_cases(self, test_data_file, table_column_types: TableColumnTypes):
        if test_data_file.endswith(".xlsx"):
            cases = TestDataFile(test_data_file, sql_reader=self.sql_reader, backend=self.backend).parse_test_cases(
                table_column_types
            )
        elif test_data_file.endswith(".json"):
            with open(test_data_file, "r") as f:
                cases = json.loads(f.read(), object_hook=json_util.object_hook)
            cases = [TestCase.from_dict(case_dict) for case_dict in cases]
        else:
            raise AssertionError(f"unsupported format of test file: {test_data_file}")
        return cases

    def run_case(self, case: TestCase) -> bool:
        try:
            self.test_case_runner.run_test(case)
            return True
        except Exception:
            import traceback

            traceback.print_exc()
            return False

    def convert_cases_to_json(self, test_data_file: str):
        if not test_data_file.endswith(".xlsx"):
            raise AssertionError(f"only support to convert excel file, got `{test_data_file}`")
        cases = TestDataFile(test_data_file, sql_reader=self.sql_reader).parse_test_cases(self.table_column_types)
        data = [case.as_dict() for case in cases]
        output_file = test_data_file[: -len(".xlsx")] + ".json"
        with open(output_file, "w") as f:
            f.write(json.dumps(data, default=json_util.default, ensure_ascii=False, indent=2, sort_keys=True))
            logger.info(f"created file: {output_file}")

    def generate_python_unittest_file(self, test_data_file: str, backend: str = "spark"):
        cases = self.parse_test_cases(test_data_file, self.table_column_types)
        py_file = os.path.join(
            os.path.dirname(test_data_file),
            os.path.basename(test_data_file).replace(".", "__").replace("__xlsx", "_test.py"),
        )
        import jinja2

        env = jinja2.Environment(
            loader=jinja2.DictLoader(
                {
                    "py_file_tpl": f"""import os
import unittest

from {self.__module__} import SqlTester


class SqlTest(unittest.TestCase):

    def __init__(self, methodName='runTest'):
        super().__init__(methodName)
        work_dir = os.environ.get('WORK_DIR')
        self.sql_tester = SqlTester(unit_test_case=self, work_dir=work_dir, backend='{backend}')
        this_file = os.path.abspath(__file__)
        self.test_data_file = os.path.join(os.path.dirname(this_file), os.path.basename(this_file).replace('__', '.').replace('_test.py', '.json'))
        if not os.path.isfile(self.test_data_file):
            self.test_data_file = os.path.join(os.path.dirname(this_file), os.path.basename(this_file).replace('__', '.').replace('_test.py', '.xlsx'))
{{% for case in cases %}}
    def test_{{{{loop.index}}}}(self):
        # {{{{ case.name }}}}
        self.assertTrue(self.sql_tester.run_test(self.test_data_file, {{{{loop.index0}}}}).is_success)

{{% endfor %}}
if __name__ == '__main__':
    unittest.main()
"""
                }
            )
        )
        with open(py_file, "wb") as f:
            env.get_template("py_file_tpl").stream(cases=cases).dump(f, encoding="utf8")
            logger.info(f"created file: {py_file}")


class SqlReader:
    def read_sql(self, test_data_file: str) -> str:
        raise NotImplementedError()

    def find_file_path(self, file_name: str) -> str:
        file_name = os.path.basename(file_name)
        for root, _dirs, files in os.walk(work_path.work_path()):
            for file in files:
                if file == file_name:
                    return work_path.relative_path(os.path.join(root, file_name))

    def read_as_content(self, test_data_file: str):
        return False

    def read_as_file_path(self, test_data_file: str):
        return not self.read_as_content(test_data_file)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['easy_sql',
 'easy_sql.sql_linter',
 'easy_sql.sql_linter.rules',
 'easy_sql.sql_processor',
 'easy_sql.sql_processor.backend',
 'easy_sql.sql_processor.backend.sql_dialect',
 'easy_sql.udf',
 'easy_sql.utils']

package_data = \
{'': ['*']}

extras_require = \
{'cli': ['click>=8.1.3,<9.0.0'],
 'clickhouse': ['SQLAlchemy>=1.4.40,<2.0.0',
                'clickhouse-driver>=0.2.4,<0.3.0',
                'clickhouse-sqlalchemy>=0.2.1,<0.3.0'],
 'linter': ['regex>=2022.7.25,<2023.0.0',
            'colorlog>=6.6.0,<7.0.0',
            'sqlfluff>=1.2.1,<2.0.0'],
 'pg': ['SQLAlchemy>=1.4.40,<2.0.0', 'psycopg2-binary>=2.9.3,<3.0.0'],
 'spark': ['pyspark>=3.3.0,<4.0.0']}

setup_kwargs = {
    'name': 'easy-sql-easy-sql-j',
    'version': '0.3.0',
    'description': 'A library developed to ease the data ETL development process.',
    'long_description': '# Easy SQL\n\nEasy SQL is built to ease the data ETL development process.\nWith Easy SQL, you can develop your ETL in SQL in an imperative way.\nIt defines a few simple syntax on top of standard SQL, with which SQL could be executed one by one.\nEasy SQL also provides a processor to handle all the new syntax.\nSince this is SQL agnostic, any SQL engine could be plugged-in as a backend.\nThere are built-in support for several popular SQL engines, including SparkSQL, PostgreSQL, Clickhouse, Aliyun Maxcompute, Google BigQuery.\nMore will be added in the near future.\n\n- Docs: <https://easy-sql.readthedocs.io/>\n- Enterprise extended product: <https://data-workbench.com/>\n\n[![GitHub Action Build](https://github.com/easysql/easy_sql/actions/workflows/build.yaml/badge.svg?branch=main&event=push)](https://github.com/easysql/easy_sql/actions/workflows/build.yaml?query=branch%3Amain+event%3Apush)\n[![Docs Build](https://readthedocs.org/projects/easy-sql/badge/?version=latest)](https://easy-sql.readthedocs.io/en/latest/?badge=latest)\n[![EasySQL Coverage](https://codecov.io/gh/easysql/easy_sql/branch/main/graph/badge.svg)](https://codecov.io/gh/easysql/easy_sql)\n\n## Install Easy SQL\n\nInstall Easy SQL using pip: `python3 -m pip install easy_sql-easy_sql`\n\n## Building Easy SQL\n\nEasy SQL is developed in Python and could be built with the following make command:\n\n```bash\nmake package-pip\n```\n\nAfter the above command, there will be a file named `easy_sql*.whl` generated in the `dist` folder.\nYou can install it with command `pip install dist/easy_sql*.whl`.\n\n## Dependencies\n\nSince there are several backends, we only need to install some specific dependencies if we only use one of them.\n\nFor spark, you need to install some version of `pyspark`.\n\nFor other backends, install the dependencies as listed below:\n```\n# for pg/clickhouse/bigquery backend only\nSQLAlchemy==1.3.23\n# for pg backend only\npsycopg2-binary==2.8.6\n# for clickhouse backend only\nclickhouse-driver==0.2.0\nclickhouse-sqlalchemy==0.1.6\n# for BigQuery backend only\nsqlalchemy-bigquery==1.4.3\n# for MaxCompute backend only\npyodps==0.10.7.1\n```\n\nTo use other tools / functionalities, install the dependencies as listed below:\n\n```\n# for Linter only\nsqlfluff==1.2.1\ncolorlog==4.0.2\nregex==2022.6.2\n# for command line tools only\nclick==6.7\n```\n\n## First ETL with Easy SQL\n\n(You need to install click package (by command `python3 -m pip install click==6.7`) before run the command below.)\n\n### For spark backend\n\nCreate a file named `sample_etl.spark.sql` with content as below:\n\n```sql\n-- prepare-sql: drop database if exists sample cascade\n-- prepare-sql: create database sample\n-- prepare-sql: create table sample.test as select 1 as id, \'1\' as val\n\n-- target=variables\nselect true as __create_output_table__\n\n-- target=variables\nselect 1 as a\n\n-- target=log.a\nselect \'${a}\' as a\n\n-- target=log.test_log\nselect 1 as some_log\n\n-- target=check.should_equal\nselect 1 as actual, 1 as expected\n\n-- target=temp.result\nselect\n    ${a} as id, ${a} + 1 as val\nunion all\nselect id, val from sample.test\n\n-- target=output.sample.result\nselect * from result\n\n-- target=log.sample_result\nselect * from sample.result\n```\n\nRun it with command:\n\n```bash\nbash -c "$(python3 -m easy_sql.data_process -f sample_etl.spark.sql -p)"\n```\n\n### For postgres backend:\n\nYou need to start a postgres instance first.\n\nIf you have docker, run the command below:\n\n```bash\ndocker run -d --name postgres -p 5432:5432 -e POSTGRES_PASSWORD=123456 postgres\n```\n\nCreate a file named `sample_etl.postgres.sql` with content as the test file [here](https://github.com/easysql/easy_sql/blob/main/test/sample_etl.postgres.sql).\n\nRun it with command:\n\n```bash\nPG_URL=postgresql://postgres:123456@localhost:5432/postgres python3 -m easy_sql.data_process -f sample_etl.postgres.sql\n```\n\n### For clickhouse backend:\n\nYou need to start a clickhouse instance first.\n\nIf you have docker, run the command below:\n\n```bash\ndocker run -d --name clickhouse -p 9000:9000 yandex/clickhouse-server:20.12.5.18\n```\n\nCreate a file named `sample_etl.clickhouse.sql` with content as the test file [here](https://github.com/easysql/easy_sql/blob/main/test/sample_etl.clickhouse.sql).\n\nRun it with command:\n\n```bash\nCLICKHOUSE_URL=clickhouse+native://default@localhost:9000 python3 -m easy_sql.data_process -f sample_etl.clickhouse.sql\n```\n\n### For other backends:\n\nThe usage is similar, please refer to API.\n\n## Run ETL in your code\n\nEasy SQL can be used as a very light-weight library. If you\'d like to run ETL programmatically in your code.\nPlease refer to the code snippets below:\n\n```python\nfrom pyspark.sql import SparkSession\n\nfrom easy_sql.sql_processor import SqlProcessor\nfrom easy_sql.sql_processor.backend import SparkBackend\n\nif __name__ == \'__main__\':\n    spark = SparkSession.builder.enableHiveSupport().getOrCreate()\n    backend = SparkBackend(spark)\n    sql = \'\'\'\n-- target=log.some_log\nselect 1 as a\n    \'\'\'\n    sql_processor = SqlProcessor(backend, sql)\n    sql_processor.run()\n```\n\nMore sample code about other backends could be referred [here](https://github.com/easysql/easy_sql/blob/main/test/sample_data_process.py)\n\n## Debugging ETL\n\nWe recommend debugging ETLs from jupyter. You can follow the steps below to start debugging your ETL.\n\n1. Install jupyter first with command `pip install jupyterlab`.\n\n2. Create a file named `debugger.py` with contents like below:\n\nA more detailed sample could be found [here](https://github.com/easysql/easy_sql/blob/main/debugger.py).\n\n```python\nfrom typing import Dict, Any\n\ndef create_debugger(sql_file_path: str, vars: Dict[str, Any] = None, funcs: Dict[str, Any] = None):\n    from pyspark.sql import SparkSession\n    from easy_sql.sql_processor.backend import SparkBackend\n    from easy_sql.sql_processor_debugger import SqlProcessorDebugger\n    spark = SparkSession.builder.enableHiveSupport().getOrCreate()\n    backend = SparkBackend(spark)\n    debugger = SqlProcessorDebugger(sql_file_path, backend, vars, funcs)\n    return debugger\n\n```\n\n3. Create a file named `test.sql` with contents as [here](https://github.com/easysql/easy_sql/blob/main/test/sample_etl.spark.sql).\n\n4. Then start jupyter lab with command: `jupyter lab`.\n\n5. Start debugging like below:\n\n![ETL Debugging](https://raw.githubusercontent.com/easysql/easy_sql/main/debugger-usage.gif)\n\n## Contributing\n\nPlease submit PR.\n',
    'author': 'Easy SQL from Thoughtworks',
    'author_email': 'easy_sql@thoughtworks.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://easy-sql.readthedocs.io',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Dict, List, Tuple

from easy_sql.sql_processor.backend.sql_dialect import SqlDialect, SqlExpr

if TYPE_CHECKING:
    from sqlalchemy.engine import ResultProxy

    from ..base import Partition

__all__ = ["ChSqlDialect"]

from ...common import SqlProcessorAssertionError


def split_table_name(table_name: str) -> Tuple[str, ...]:
    if len(table_name.split(".")) != 2:
        raise Exception(f"cannot split table name: {table_name}")
    return tuple(table_name.split("."))


class ChSqlDialect(SqlDialect):
    def __init__(self, sql_expr: SqlExpr, partitions_table_name: str):
        self.partitions_table_name = partitions_table_name
        super(ChSqlDialect, self).__init__(sql_expr)

    def create_db_sql(self, db: str):
        return f"create database if not exists {db}"

    def use_db_sql(self, db: str):
        return f"use {db}"

    def drop_db_sql(self, db: str) -> List[str]:
        drop_db = f"drop database if exists {db}"
        drop_pt_metadata = f"alter table {self.partitions_table_name} delete where db_name = '{db}'"
        return [drop_db, drop_pt_metadata]

    def rename_table_sql(self, from_table: str, to_table: str) -> str:
        return f"rename table {from_table} to {to_table}"

    def rename_table_db_sql(self, table_name: str, schema: str):
        pure_table_name = table_name if "." not in table_name else table_name[table_name.index(".") + 1 :]
        return f"rename table {table_name} to {schema}.{pure_table_name}"

    def rename_view_sql(self, from_table: str, to_table: str) -> str:
        return self.rename_table_sql(from_table, to_table)

    def drop_view_sql(self, table: str) -> str:
        return f"drop table {table}"

    def create_view_sql(self, table_name: str, select_sql: str) -> str:
        return f"create view {table_name} as {select_sql}"

    def get_tables_sql(self, db: str) -> str:
        return f"show tables in {db}"

    def get_dbs_sql(self) -> str:
        return "show databases"

    def create_table_sql(self, table_name: str, select_sql: str):
        return f"create table {table_name} engine=MergeTree order by tuple() as {select_sql}"

    def support_native_partition(self) -> bool:
        return True

    def support_move_individual_partition(self) -> bool:
        return True

    def delete_partition_sql(self, table_name, partitions: List[Partition]) -> List[str]:
        if len(partitions) > 1:
            raise SqlProcessorAssertionError(
                f"Only support exactly one partition column, found: {[str(p) for p in partitions]}"
            )
        db, pure_table_name = split_table_name(table_name)

        drop_pt_metadata = (
            f"alter table {self.partitions_table_name} delete "
            f"where db_name = '{db}' and table_name = '{pure_table_name}' "
            f"and partition_value = '{partitions[0].value}'"
        )

        pt_expr = f'tuple({", ".join([self.sql_expr.for_value(pt.value) for pt in partitions])})'
        drop_pt = f"alter table {table_name} drop partition {pt_expr}"

        return [drop_pt, drop_pt_metadata]

    def native_partitions_sql(self, table_name: str) -> Tuple[str, Callable[[ResultProxy], List[str]]]:
        return f"show create table {table_name}", self.extract_partition_cols

    def extract_partition_cols(self, native_partitions_sql_result: ResultProxy) -> List[str]:
        create_table_sql_lines = native_partitions_sql_result.fetchall()[0][0].split("\n")
        pt_cols = []
        for line in create_table_sql_lines:
            if line.startswith("PARTITION BY ("):
                pt_cols = [col.strip() for col in line[len("PARTITION BY (") : -1].split(",")]
                break
            elif line.startswith("PARTITION BY "):
                pt_cols = [line[len("PARTITION BY ") :].strip()]
                break
        return pt_cols

    def create_table_with_partitions_sql(self, table_name: str, cols: List[Dict], partitions: List[Partition]):
        cols_expr = ",\n".join([f"{col['name']} {col['type']}" for col in cols])
        if len(partitions) == 0:
            partition_expr = ""
        elif len(partitions) == 1:
            partition_expr = f"partition by {partitions[0].field}"
        else:
            partition_expr = f'partition by tuple({", ".join([pt.field for pt in partitions])})'
        return (
            f"create table if not exists {table_name} (\n{cols_expr}\n)\nengine=MergeTree\n{partition_expr}\norder by"
            " tuple() settings allow_nullable_key=1;"
        )

    def create_temp_table_with_schema_sql(self, temp_table_name: str):
        return f'create table if not exists {temp_table_name} as {temp_table_name.split("__temp")[0]}'

    def create_partition_automatically(self):
        return True

    def create_temp_table_schema_as_target(self):
        return True

    def insert_data_sql(
        self, table_name: str, col_names_expr: str, select_sql: str, partitions: List[Partition]
    ) -> List[str]:
        insert_data_sql = f"insert into {table_name}({col_names_expr}) {select_sql}"
        self._check_no_none_in_partition_values(partitions)
        if len(partitions) != 0:
            self._check_partition_field_single(partitions)
            drop_pt_metadata_if_exist, insert_pt_metadata = self._generate_pt_metadata_sql(table_name, partitions)
            return [insert_data_sql, drop_pt_metadata_if_exist, insert_pt_metadata]
        return [insert_data_sql]

    def move_data_sql(self, target_table_name: str, temp_table_name: str, partitions: List[Partition]):
        self._check_no_none_in_partition_values(partitions)
        if len(partitions) != 0:
            self._check_partition_field_single(partitions)
            drop_pt_metadata_if_exist, insert_pt_metadata = self._generate_pt_metadata_sql(
                target_table_name, partitions
            )
            # Assumption: partition movement requires the two tables to have
            # the same structure, partition key, engine family and storage policy,
            # need to upgrade local clickhouse version (> 21) to
            # support moving feature https://github.com/ClickHouse/ClickHouse/issues/14582
            move_partition_sqls = []
            for partition in partitions:
                move_partition_sqls.append(
                    f"alter table {temp_table_name} move partition '{partition.value}' to table {target_table_name}"
                )
            move_data_sql = " ".join(move_partition_sqls)
            return [move_data_sql, drop_pt_metadata_if_exist, insert_pt_metadata]

    def drop_table_sql(self, table: str) -> List[str]:
        drop_table_sql = f"drop table if exists {table}"
        db, pure_table_name = split_table_name(table)
        drop_pt_metadata = (
            f"alter table {self.partitions_table_name} delete where db_name = '{db}' and table_name ="
            f" '{pure_table_name}'"
        )
        return [drop_table_sql, drop_pt_metadata]

    def support_static_partition(self):
        return False

    def create_pt_meta_table_sql(self, db: str) -> str:
        # As clickhouse has create partition table in RdbBackend, no need to create again
        return "select 1"

    def insert_pt_metadata_sql(self, table_name: str, partitions: List[Partition]) -> str:
        if len(partitions) == 0:
            return ""
        elif len(partitions) > 1:
            raise SqlProcessorAssertionError("clickhouse only supports single-column partitioning.")
        else:
            db, pure_table_name = tuple(table_name.split("."))
            insert_pt_metadata = (
                f"insert into {self.partitions_table_name} values('{db}', '{pure_table_name}', '{partitions[0].value}',"
                " now());"
            )
            return insert_pt_metadata

    def create_table_like_sql(self, target_table_name: str, source_table_name: str, partitions: List[Partition]) -> str:
        return f"create table if not exists {target_table_name} as {source_table_name}"

    def _generate_pt_metadata_sql(self, table_name: str, partitions: List[Partition]) -> List[str]:
        db, pure_table_name = split_table_name(table_name)
        drop_pt_metadata_if_exist = (
            f"alter table {self.partitions_table_name} delete "
            f"where db_name = '{db}' and table_name = '{pure_table_name}' "
            f"and partition_value = '{partitions[0].value}'"
        )
        insert_pt_metadata = self.insert_pt_metadata_sql(table_name, partitions)
        return drop_pt_metadata_if_exist, insert_pt_metadata

    def _check_no_none_in_partition_values(self, partitions: List[Partition]):
        if any([pt.value is None for pt in partitions]):
            raise SqlProcessorAssertionError(
                f"cannot insert data when partition value is None, partitions: {partitions}, there maybe some bug,"
                " please check"
            )

    def _check_partition_field_single(self, partitions: List[Partition]):
        if len(partitions) > 1:
            raise SqlProcessorAssertionError(
                "for now clickhouse backend only support table with single field partition"
            )



from typing import Callable, Dict

from .backend import Backend, SparkBackend
from .backend.rdb import RdbBackend
from .common import SqlProcessorException, VarsReplacer

__all__ = ["FuncRunner"]


class FuncRunner:
    _instance = None

    def __init__(self, funcs: Dict[str, Callable] = None):
        self.funcs: Dict[str, Callable] = funcs or {}

    def register_funcs(self, funcs: Dict[str, Callable]):
        self.funcs.update(funcs)

    @staticmethod
    def create(backend: Backend) -> "FuncRunner":
        import builtins

        builtin_funcs = {
            func: getattr(builtins, func)
            for func in dir(builtins)
            if not func.startswith("_") and func[0].islower() and callable(getattr(builtins, func))
        }

        import operator

        operator_funcs = {
            func: getattr(operator, func)
            for func in dir(operator)
            if not func.startswith("_") and func[0].islower() and callable(getattr(operator, func))
        }
        operator_funcs.update({"equal": operator.eq})
        all_funcs = {}
        all_funcs.update(builtin_funcs)
        all_funcs.update(operator_funcs)
        all_funcs.update(
            {
                "is_greater_or_equal": lambda a, b: a >= b,
                "equal_ignore_case": lambda a, b: a.lower() == b.lower(),
            }
        )

        if isinstance(backend, (SparkBackend,)):
            all_funcs.update(FuncRunner._get_spark_funcs(backend))
        elif isinstance(backend, (RdbBackend,)):
            all_funcs.update(FuncRunner._get_rdb_funcs(backend))

        FuncRunner._instance = FuncRunner(all_funcs)
        return FuncRunner._instance

    @staticmethod
    def _get_rdb_funcs(backend) -> Dict[str, Callable]:
        from easy_sql.sql_processor.funcs_rdb import (
            ColumnFuncs,
            ModelFuncs,
            PartitionFuncs,
            TableFuncs,
        )

        partition_funcs = PartitionFuncs(backend)
        col_funcs = ColumnFuncs(backend)
        table_funcs = TableFuncs(backend)
        model_funcs = ModelFuncs(backend)
        return {
            "partition_exists": partition_funcs.partition_exists,
            "partition_not_exists": partition_funcs.partition_not_exists,
            "is_first_partition": partition_funcs.is_first_partition,
            "is_not_first_partition": partition_funcs.is_not_first_partition,
            "previous_partition_exists": partition_funcs.previous_partition_exists,
            "get_partition_or_first_partition": partition_funcs.get_partition_or_first_partition,
            "ensure_dwd_partition_exists": partition_funcs.ensure_dwd_partition_exists,
            "get_partition_col": partition_funcs.get_partition_col,
            "get_first_partition": partition_funcs.get_first_partition,
            "get_last_partition": partition_funcs.get_last_partition,
            "ensure_partition_exists": partition_funcs.ensure_partition_exists,
            "ensure_partition_or_first_partition_exists": partition_funcs.ensure_partition_or_first_partition_exists,
            "all_cols_without_one_expr": col_funcs.all_cols_without_one_expr,
            "all_cols_with_exclusion_expr": col_funcs.all_cols_with_exclusion_expr,
            "bq_model_predict_with_local_spark": model_funcs.bq_model_predict_with_local_spark,
            "model_predict_with_local_spark": model_funcs.model_predict_with_local_spark,
            "ensure_no_null_data_in_table": table_funcs.ensure_no_null_data_in_table,
            "check_not_null_column_in_table": table_funcs.check_not_null_column_in_table,
            "all_cols_prefixed_with_exclusion_expr": col_funcs.all_cols_prefixed_with_exclusion_expr,
        }

    @staticmethod
    def _get_spark_funcs(backend) -> Dict[str, Callable]:
        from easy_sql.sql_processor.funcs_spark import (
            CacheFuncs,
            ColumnFuncs,
            IOFuncs,
            ModelFuncs,
            ParallelismFuncs,
            PartitionFuncs,
            TableFuncs,
        )

        spark = backend.spark
        partition_funcs = PartitionFuncs(backend)
        parallelism_funcs = ParallelismFuncs(spark)
        cache_funcs = CacheFuncs(spark)
        col_funcs = ColumnFuncs(backend)
        table_funcs = TableFuncs(backend)
        io_funcs = IOFuncs(spark)
        model_funcs = ModelFuncs(spark)
        return {
            "repartition": parallelism_funcs.repartition,
            "repartition_by_column": parallelism_funcs.repartition_by_column,
            "coalesce": parallelism_funcs.coalesce,
            "set_shuffle_partitions": parallelism_funcs.set_shuffle_partitions,
            "partition_exists": partition_funcs.partition_exists,
            "partition_not_exists": partition_funcs.partition_not_exists,
            "is_first_partition": partition_funcs.is_first_partition,
            "is_not_first_partition": partition_funcs.is_not_first_partition,
            "previous_partition_exists": partition_funcs.previous_partition_exists,
            "get_partition_or_first_partition": partition_funcs.get_partition_or_first_partition,
            "ensure_dwd_partition_exists": partition_funcs.ensure_dwd_partition_exists,
            "get_partition_col": partition_funcs.get_partition_col,
            "get_first_partition": partition_funcs.get_first_partition,
            "get_last_partition": partition_funcs.get_last_partition,
            "ensure_partition_exists": partition_funcs.ensure_partition_exists,
            "ensure_partition_or_first_partition_exists": partition_funcs.ensure_partition_or_first_partition_exists,
            "all_cols_without_one_expr": col_funcs.all_cols_without_one_expr,
            "all_cols_with_exclusion_expr": col_funcs.all_cols_with_exclusion_expr,
            "all_cols_prefixed_with_exclusion_expr": col_funcs.all_cols_prefixed_with_exclusion_expr,
            "ensure_no_null_data_in_table": table_funcs.ensure_no_null_data_in_table,
            "check_not_null_column_in_table": table_funcs.check_not_null_column_in_table,
            "unpersist": cache_funcs.unpersist,
            "write_csv": io_funcs.write_csv,
            "rename_csv_output": io_funcs.rename_csv_output,
            "write_json_local": io_funcs.write_json_local,
            "update_json_local": io_funcs.update_json_local,
            "model_predict": model_funcs.model_predict,
        }

    def run_func(self, func_def: str, vars_replacer: VarsReplacer) -> bool:
        func_name = func_def[: func_def.index("(")]
        if func_name not in self.funcs:
            raise SqlProcessorException(f"no function found for {func_def} in sql_processor: {func_def}")
        func = self.funcs[func_name]
        try:
            original_params = func_def[func_def.index("(") + 1 : func_def.index(")")].strip()
        except ValueError:
            raise SqlProcessorException("parse params failed for func definition: " + func_def)
        params = []
        if original_params:
            params = original_params.split(",")
            params = [vars_replacer.replace_variables(p.strip(), False) for p in params]
        return func(*params)



import json
from datetime import datetime
from typing import Any


class EsService:
    def __init__(self, base_url: str, should_send: bool = True):
        self.base_url = base_url.strip("/")
        self.should_send = should_send
        self.data = None

    def post(self, url_path: str, data: str):
        import requests

        if self.should_send:
            resp = requests.post(
                self.base_url + url_path, headers={"Content-Type": "Application/json"}, data=data.encode("utf8")
            )
            if not resp.ok:
                raise Exception(f"send data quality report failed(status={resp.status_code}): {resp.text}")
            print("data post to es done")
        else:
            self.data = {"method": "post", "args": {"url_path": url_path, "data": data}}
            print("will not send data")

    def put(self, url_path: str, data: str):
        import requests

        if self.should_send:
            resp = requests.put(
                self.base_url + url_path, headers={"Content-Type": "Application/json"}, data=data.encode("utf8")
            )
            if not resp.ok:
                raise Exception(f"send data quality report failed(status={resp.status_code}): {resp.text}")
            print("data put to es done")
        else:
            self.data = {"method": "put", "args": {"url_path": url_path, "data": data}}
            print("will not send data")

    def delete_by_query(self, index: str, query: object):
        import requests

        data = json.dumps({"query": query})
        url_path = f"/{index}/_delete_by_query"
        if self.should_send:
            resp = requests.post(
                self.base_url + url_path, headers={"Content-Type": "Application/json"}, data=data.encode("utf8")
            )
            if not resp.ok:
                raise Exception(f"send data quality report failed(status={resp.status_code}): {resp.text}")
        else:
            self.data = {"method": "post", "args": {"url_path": url_path, "data": data}}
            print("will not send data")


class Reporter:
    def __init__(self, es_service: EsService, index_prefix: str = "", now: Any = None):
        self.es_service = es_service
        self.now = now
        self.index_prefix = index_prefix

    def _es_index_name(self, name: str):
        return f"{self.index_prefix}_{name}" if self.index_prefix else name

    def report_task_result(self, task_id: str, report: str):
        """
        es index:

        PUT /{index_prefix}_task_report
        {
          "mappings": {
            "properties": {
              "task_id": { "type": "wildcard" },
              "report": { "type": "text" },
              "created_at": { "type": "date", "format": "yyyy-MM-dd HH:mm:ss" }
            }
          }
        }
        """
        now = (self.now or datetime.utcnow()).strftime("%Y-%m-%d %H:%M:%S")
        data = {"task_id": task_id, "report": report, "created_at": now}
        self.es_service.post(f'/{self._es_index_name("task_report")}/_doc', json.dumps(data))



from __future__ import annotations

import os
import uuid
from datetime import datetime
from typing import TYPE_CHECKING, Dict, List

if TYPE_CHECKING:
    from pyspark.sql import SparkSession

from easy_sql.sql_processor import SqlProcessor
from easy_sql.sql_processor.backend.rdb import SqlExpr

from .local_spark import LocalSpark
from .logger import log_time


def should_run_integration_test(key: str = None):
    if key is None or key in ["pg", "ch", "mc", "bq"]:
        return False
    return False


TEST_PG_URL = os.environ.get("PG_URL", "postgresql://postgres:123456@testpg:15432/postgres")
TEST_CH_URL = os.environ.get("CLICKHOUSE_URL", "clickhouse+native://default@testch:30123")
TEST_BQ_URL = os.environ.get("BQ_URL", "bigquery://")

__partition_col_converter__ = (
    lambda col: f"PARSE_DATE('%Y-%m', {col}) as {col}"
    if col in ["data_month", ":data_month"]
    else f"CAST({col} as DATE)"
)
__partition_value_converter__ = (
    lambda col, value: datetime.strptime(value, "%Y-%m").date()
    if col == "data_month"
    else datetime.strptime(value, "%Y-%m-%d").date()
)
__column_sql_type_converter__ = (
    lambda backend_type, col_name, col_type: "DATE" if col_name in ["di", "dt", "data_date", "data_month"] else None
)
__partition_expr__ = (
    lambda backend_type, partition_col: f"DATE_TRUNC({partition_col}, MONTH)"
    if backend_type == "bigqiery" and partition_col == "data_month"
    else partition_col
)
bigquery_sql_expr = SqlExpr(
    column_sql_type_converter=__column_sql_type_converter__,
    partition_col_converter=__partition_col_converter__,
    partition_value_converter=__partition_value_converter__,
    partition_expr=__partition_expr__,
)


def dt(dt_s):
    return datetime.strptime(dt_s, "%Y-%m-%d %H:%M:%S")


def date(s):
    return datetime.strptime(s, "%Y-%m-%d").date()


def dt_zone(dt_s: str, formate="%Y-%m-%d %H:%M:%S", timezone=None):
    if timezone is None:
        return datetime.strptime(dt_s, formate)
    else:
        return datetime.strptime(dt_s, formate).replace(tzinfo=timezone)


def next_id():
    return str(uuid.uuid1()).replace("-", "")


@log_time
def run_sql(
    sql: str,
    result_table: str,
    funcs: dict = None,
    variables: dict = None,
    dry_run: bool = False,
    spark: SparkSession = None,
    spark_conf: Dict = None,
) -> List:
    spark = spark or LocalSpark.get(spark_conf)
    processor = SqlProcessor(spark, sql, [], variables or {})
    processor.func_runner.register_funcs(funcs or {})
    processor.run(dry_run=dry_run)
    return spark.sql(f"select * from {result_table}").collect()



import os.path
import unittest

from easy_sql import data_process

proj_base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


class DataProcessTest(unittest.TestCase):
    def test_spark(self):
        command = data_process._data_process(os.path.join(proj_base_dir, "test/sample_etl.spark.sql"), None, None, True)
        self.assertRegex(
            command,
            r"spark-submit --conf spark.master=local\[2\] --conf spark.submit.deployMode=client "
            r"--conf spark.app.name=sample_etl.spark_[\d]+ "
            "--conf spark.sql.warehouse.dir=/tmp/spark-warehouse-localdw "
            '--conf spark.driver.extraJavaOptions="-Dderby.system.home=/tmp/spark-warehouse-metastore '
            '-Dderby.stream.error.file=/tmp/spark-warehouse-metastore.log" '
            '--conf spark.files="[^"]+test/sample_etl.spark.sql" '
            '"[^"]+/easy_sql/data_process.py" '
            "-f .+/test/sample_etl.spark.sql --dry-run 0",
        )
        data_process._data_process(os.path.join(proj_base_dir, "test/sample_etl.spark.sql"), None, None, False)

    def test_postgres(self):
        data_process._data_process(os.path.join(proj_base_dir, "test/sample_etl.postgres.sql"), None, None, False)

    def test_clickhouse(self):
        data_process._data_process(os.path.join(proj_base_dir, "test/sample_etl.clickhouse.sql"), None, None, False)
